/*
**    TP CPE Lyon
**    Copyright (C) 2014 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mesh_opengl.hpp"

#include "../libmesh/mesh.hpp"
#include "../local/glutils.hpp"
#include "../libcommon/exception_cpe.hpp"



namespace cpe
{

mesh_opengl::mesh_opengl()
    :vbo_vertex(0),vbo_normal(0),vbo_color(0),vbo_index(0),number_of_triangles(0)
{

}

mesh_opengl::~mesh_opengl()
{
    delete_vbo();
}

void mesh_opengl::fill_vbo(const mesh& m)
{
    PRINT_OPENGL_ERROR();

    //number of vertices
    const int N_vertex=m.number_of_vertices();
    if(N_vertex<=0)
        throw cpe::exception_cpe("incorrect number of vertices",MACRO_EXCEPTION_PARAMETER);

    number_of_triangles=m.number_of_triangle();
    if(number_of_triangles<=0)
        throw cpe::exception_cpe("incorrect number of triangles",MACRO_EXCEPTION_PARAMETER);

    //array of vertex
    const std::vector<vec3>& vertices=m.get_vertices();
    if(static_cast<int>(vertices.size())!=N_vertex)
        throw cpe::exception_cpe("incorrect size of vertices",MACRO_EXCEPTION_PARAMETER);

    //array of normals
    const std::vector<vec3>& normal=m.get_normal();
    if(static_cast<int>(normal.size())!=N_vertex)
        throw cpe::exception_cpe("Error: normals should have the same size than vertices. Are the normals computed?",MACRO_EXCEPTION_PARAMETER);

    //array of colors
    const std::vector<vec3>& color=m.get_color();
    if(static_cast<int>(color.size())!=N_vertex)
        throw cpe::exception_cpe("Error: color should have the same size than vertices. Are the color set up?",MACRO_EXCEPTION_PARAMETER);


    //array of indices
    const std::vector<int>& index=m.get_connectivity();
    if(index.size()!=number_of_triangles*3)
        throw cpe::exception_cpe("Error: connectivity size is incorrect.",MACRO_EXCEPTION_PARAMETER);


    //get all the pointers
    const float* ptr_vertices=vertices[0].pointer();
    const float* ptr_normals=normal[0].pointer();
    const float* ptr_colors=color[0].pointer();
    const int* ptr_indices=&index[0];


    //create the new vbo
    if(vbo_vertex==0)
        glGenBuffers(1,&vbo_vertex);PRINT_OPENGL_ERROR();
    if(vbo_normal==0)
        glGenBuffers(1,&vbo_normal);PRINT_OPENGL_ERROR();
    if(vbo_color==0)
        glGenBuffers(1,&vbo_color);PRINT_OPENGL_ERROR();
    glGenBuffers(1,&vbo_index);PRINT_OPENGL_ERROR();


    //VBO vertex
    glBindBuffer(GL_ARRAY_BUFFER,vbo_vertex); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_vertex))
        throw cpe::exception_cpe("vbo_buffer incorrect",MACRO_EXCEPTION_PARAMETER);
    glBufferData(GL_ARRAY_BUFFER,3*sizeof(float)*N_vertex,ptr_vertices,GL_DYNAMIC_DRAW);
    PRINT_OPENGL_ERROR();

    //VBO normal
    glBindBuffer(GL_ARRAY_BUFFER,vbo_normal); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_normal))
        throw cpe::exception_cpe("vbo_normal incorrect",MACRO_EXCEPTION_PARAMETER);
    glBufferData(GL_ARRAY_BUFFER,3*sizeof(float)*N_vertex,ptr_normals,GL_DYNAMIC_DRAW);
    PRINT_OPENGL_ERROR();

    //VBO color
    glBindBuffer(GL_ARRAY_BUFFER,vbo_color); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_color))
        throw cpe::exception_cpe("vbo_color incorrect",MACRO_EXCEPTION_PARAMETER);
    glBufferData(GL_ARRAY_BUFFER,3*sizeof(float)*N_vertex,ptr_colors,GL_DYNAMIC_DRAW);
    PRINT_OPENGL_ERROR();


    //VBO index
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,vbo_index); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_index))
        throw cpe::exception_cpe("vbo_index incorrect",MACRO_EXCEPTION_PARAMETER);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER,3*sizeof(int)*number_of_triangles,ptr_indices,GL_DYNAMIC_DRAW);
    PRINT_OPENGL_ERROR();


    PRINT_OPENGL_ERROR();
}

void mesh_opengl::delete_vbo()
{
    PRINT_OPENGL_ERROR();

    if(vbo_vertex!=0)
        glDeleteBuffers(1,&vbo_vertex);
    PRINT_OPENGL_ERROR();

    if(vbo_normal!=0)
        glDeleteBuffers(1,&vbo_normal);
    PRINT_OPENGL_ERROR();

    if(vbo_color!=0)
        glDeleteBuffers(1,&vbo_color);
    PRINT_OPENGL_ERROR();


    if(vbo_index!=0)
        glDeleteBuffers(1,&vbo_index);
    PRINT_OPENGL_ERROR();


}

void mesh_opengl::draw()
{
    if(number_of_triangles==0)
        throw cpe::exception_cpe("incorrect number of triangles",MACRO_EXCEPTION_PARAMETER);

    PRINT_OPENGL_ERROR();

    glBindBuffer(GL_ARRAY_BUFFER,vbo_vertex); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_vertex))
        throw cpe::exception_cpe("vbo_vertex incorrect",MACRO_EXCEPTION_PARAMETER);
    glEnableClientState(GL_VERTEX_ARRAY); PRINT_OPENGL_ERROR();
    glVertexPointer(3, GL_FLOAT, 0, 0); PRINT_OPENGL_ERROR();

    glBindBuffer(GL_ARRAY_BUFFER,vbo_normal); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_normal))
        throw cpe::exception_cpe("vbo_normal incorrect",MACRO_EXCEPTION_PARAMETER);
    glEnableClientState(GL_NORMAL_ARRAY); PRINT_OPENGL_ERROR();
    glNormalPointer(GL_FLOAT, 0, 0); PRINT_OPENGL_ERROR();

    glBindBuffer(GL_ARRAY_BUFFER,vbo_color); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_color))
        throw cpe::exception_cpe("vbo_color incorrect",MACRO_EXCEPTION_PARAMETER);
    glEnableClientState(GL_COLOR_ARRAY); PRINT_OPENGL_ERROR();
    glColorPointer(3,GL_FLOAT, 0, 0); PRINT_OPENGL_ERROR();

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,vbo_index); PRINT_OPENGL_ERROR();
    if(!glIsBuffer(vbo_index))
        throw cpe::exception_cpe("vbo_index incorrect",MACRO_EXCEPTION_PARAMETER);
    glDrawElements(GL_TRIANGLES, 3*number_of_triangles, GL_UNSIGNED_INT, 0); PRINT_OPENGL_ERROR();

    PRINT_OPENGL_ERROR();
}





}


